/*! \file GeneR_glob.cc
 *  \brief Global variables of GeneR (i.e sequences buffers)
 *   The global variable myGeneR is allocated here, and access trough
 *  functions defined here too. The global variable myGeneR is a vector
 *  of elements of class GeneR_seq.
 *  \version 1
 *  \date Created       : 26/10/05 
 *  \date Last Modified : Time-stamp: <2007-10-12 19:30:38 antoine>
 *  \author A. Lucas
 *  \note Licence: CeCIll
 */

#include "GeneR_seq.h"
#include <stdio.h>
#include <vector>
#include "GeneR_glob.h"
#include "complementaire.h"




using namespace std;


// Constructor
GeneR_glob::GeneR_glob():
  max_buffers(0),
  myGeneR(0),
  strand(0)
{

}

// copy: it must not be called
GeneR_glob::GeneR_glob(const GeneR_glob & _src):
  max_buffers(0),
  myGeneR(0),
  strand(0)
{
  printf("Error line 43\n");
}

// destructor
GeneR_glob::~GeneR_glob()
{

  init(0); // we resize  vector to 0, and destruct all GeneR_seq objects
}

GeneR_glob * GeneR_glob::instance()
{
  static GeneR_glob  _instance;
  if(_instance.maxBuffers() == 0)
    {
      _instance.init(100);
    }
  return & _instance;
}

/** init_GeneR(int * maxbuffers)
 * \brief initialize all buffer, maxbuffers is a parameter, can
 *        be use to increase maximum buffers
 * \param maxbuffers number of buffers 
 */
void R_initialise_GeneR(int * maxbuffers)
{
  GeneR_glob::instance()->init( * maxbuffers);
}

void GeneR_glob::init(int  maxbuffers_p)
{

  //printf("init %d %d \n",maxbuffers_p, max_buffers);

  // on detruit les buffers a enlever
  if(maxbuffers_p < max_buffers)
    {
      for( int i =maxbuffers_p; i < max_buffers; i++)
	{
	  delete myGeneR[i];
	  myGeneR[i] = 0;
	}
    }

  // on retaille le vecteur
  myGeneR.resize(maxbuffers_p);

  // on ajoute les nouveaux
  if(maxbuffers_p > max_buffers)
    {
      for( int i =max_buffers; i < maxbuffers_p; i++)
	{
	  myGeneR[i] = new GeneR_seq();
	}
    }

  max_buffers = maxbuffers_p;

}


/** \fn myGeneRbuffer(int bufno,int comp=0)
 * \brief return the sequence of a specific buffer; NULL if bufno
 *  not in range.
 * \param bufno buffer number 
 * \param comp 0: forward, 1 reverse.
 */
char * GeneR_glob::buffer(int bufno,int comp)
{

  if((bufno <0) || (bufno > max_buffers))
    return(NULL);
  if(comp)
    {
      if(myGeneR[bufno]->getsSize() != myGeneR[bufno]->getsSizeComp())
	complementaire::buf_sys_complementaire(bufno);
      return(myGeneR[bufno]->getseqComp());

    }
  else
    return(myGeneR[bufno]->getseq());
}


/** \fn myGeneRsize(int bufno,int comp=0)
 * \brief returns size of a buffer. Returns -1 if buffer not in range
 * \param bufno buffer number 
 * \param comp 0: forward, 1 reverse.
 * \note size of buffer uninitialized is 0;
*/
int GeneR_glob::size(int bufno,int comp)
{
  if((bufno <0) || (bufno > max_buffers))
    return(-1);

  if(comp)
    return(myGeneR[bufno]->getsSizeComp());
  else
    return(myGeneR[bufno]->getsSize());
}

/** \fn myGeneRmaxBuffers()
 * \brief returns global variable max_buffers
 */
int GeneR_glob::maxBuffers()
{
  return(max_buffers);
}

/** \fn sys_maxbuffer(int * nb)
 * \brief returns global variable max_buffers
 */
void sys_maxbuffer(int * nb)
{
  *nb=GeneR_glob::instance()->maxBuffers();
}

/** \fn myGeneRallocBuffer(int size, int bufno, int comp, int fill )
 * \brief (re) allocate a buffer. fill=1: add a "\0" at position 0
 *  returns 0 if error; 1: OK.
 * \param bufno buffer number 
 * \param comp 0: forward, 1 reverse.
 * \param size size of buffer (including "\0")
 * \param fill fill=1: add a '\0' at begining of sequence
 */
int GeneR_glob::allocBuffer(int size, int bufno, int comp, int fill )
{
  if((bufno <0) || (bufno > max_buffers))
    return(0);
      
  return(myGeneR[bufno]->allocBuffer(size, comp, fill));
}




/** @brief allocate a buffer
    @param *size  Taille a allouer
    @param *bufno Numero du bufseq
    @param *fill  1: add a "\0" at begining
    @param *comp  0 : brin sens -  1 : brin anti-sens
    @param *err 0: error, 1: good
*/
/* ---------------------------------------------------- */
void myGeneR_alloc (int *size, int *bufno, int *fill, int *comp,int * err)
{
  *err =GeneR_glob::instance()->allocBuffer(*size,*bufno,*comp,*fill);
}



/** @brief size of buffer

   return length of sequence in bufno.   
   @param  *bufno Integer, number of the sequence. (bufseq)
   @param  *comp  0 : brin sens -  1 : brin anti-sens)
*/
void  size_buf (int  *bufno,int * comp)
{

  *bufno = GeneR_glob::instance()->size(*bufno,*comp);
}


/** \fn getParam(int * bufno, int * beg, int * size, int * masterSize, int * err)
 * \brief get globals parameters of a sequence
 * \param bufno buffer sequence
 * \param beg begining position on source sequence
 * \param size size of sequence (not including "\0")
 * \param masterSize size of source sequence
 * \param err 0: is buffno not in range 1 good
 */
void getParam(int * bufno, int * beg, int * size, int * masterSize, int * err)
{
  *err = GeneR_glob::instance()->getParam(* bufno, * beg,* size,*masterSize);
}

int GeneR_glob::getParam(int  bufno, int & beg, int & size, int & masterSize)
{
   if((bufno <0) || (bufno > max_buffers))
    {
      
      return 0;
    }

   beg = myGeneR[bufno]->getBegSeq();
   masterSize = myGeneR[bufno]->getMasterSize();
   size = myGeneR[bufno]->getsSize() - 1;
   return 1;
}

/** \fn setParam(int * bufno, int * beg, int * masterSize, int * err)
 * \brief set globals parameters of a sequence
 * \param bufno buffer sequence
 * \param beg begining position on source sequence
 * \param masterSize size of source sequence
 * \param err 0: is buffno not in range 1 good
 */
void setParam(int * bufno, int * beg, int * masterSize, int * err)
{
  *err =  GeneR_glob::instance()->setParam(* bufno, * beg,*masterSize);
}

int GeneR_glob::setParam(int bufno, int  beg, int  masterSize)
{
   if((bufno <0) || (bufno > max_buffers))
    {
      
      return 0;
    }
   
   myGeneR[bufno]->setBegSeq(beg);
   myGeneR[bufno]->setMasterSize(masterSize);
   /*   myGeneR[*bufno].SeqSize=*size;*/
   return 1;
}

/** \fn freeSeq(int * bufno)
 * \brief free sequence and complementaire
 * \param bufno sequence number
 * \note set bufno to 0 is error (i.e. buffer out of range)
 */
void freeSeq(int * bufno)
{
  GeneR_glob::instance()->freeSeq(*bufno);
}

/** \fn GeneRfreeSeq(int  bufno)
 * \brief free sequence and complementaire
 * \param bufno sequence number
 * \note set bufno to 0 is error (i.e. buffer out of range)
 */
void GeneR_glob::freeSeq(int  bufno)
{
  if((bufno <0) || (bufno > max_buffers))
    {
      bufno = 0;
      return;
    }
  myGeneR[bufno]->freeSeq();

}

/** \fn freeCompSeq(int  bufno)
 * \brief free sequence complementaire
 * \param bufno sequence number
 * \note set bufno to 0 is error (i.e. buffer out of range)
 */
void GeneR_glob::freeCompSeq(int  bufno)
{
  if((bufno <0) || (bufno > max_buffers))
    {
      bufno = 0;
      return;
    }
  myGeneR[bufno]->freeCompSeq();

}

/** \fn setStrand  (int * strandR)
 * \brief set global variable "strand"
 */
void setStrand (int * strandR)
{
  GeneR_glob::instance()->setstrand( strandR[0]);
}

/** \fn getStrand (int * strandR)
 * \brief get global variable "strand"
 */
void getStrand (int * strandR)
{
  strandR[0] = GeneR_glob::instance()->getstrand();
}

/** \fn  getAccno(int * bufno,char ** name,int * err)
 * \brief get accNo of a sequence
 * \param bufno number of sequence
 * \param name name of sequence
 * \param err: 0 if buffno out of range else 1
 */
void getAccno (int * bufno,char ** name,int * err)
{
  GeneR_glob::instance()->getAccno(* bufno,*name,err);
}

/** \fn  setAccno(int * bufno,char ** name,int * err)
 * \brief set accNo of a sequence
 * \param bufno number of sequence
 * \param name name of sequence
 * \param err: 0 if buffno out of range else 1
 */
void setAccno (int * bufno,char ** name,int * err)
{
  GeneR_glob::instance()->setAccno(* bufno,*name,err);
}

/** \fn  getAccno(int * bufno,char ** name,int * err)
 * \brief get accNo of a sequence
 * \param bufno number of sequence
 * \param name name of sequence
 * \param err: 0 if buffno out of range else 1
 */
void GeneR_glob::getAccno (int  bufno,char * name,int * err)
{
  if((bufno <0) || (bufno > max_buffers))
    {
      *err = 0;
      return;
    }
   strncpy(name,myGeneR[bufno]->getAccN(),MaxLNom);
   *err=1;
}

/** \fn  setAccno(int * bufno,char ** name,int * err)
 * \brief set accNo of a sequence
 * \param bufno number of sequence
 * \param name name of sequence
 * \param err: 0 if buffno out of range else 1
 */
void GeneR_glob::setAccno (int  bufno,char * name,int * err)
{
   if((bufno <0) || (bufno > max_buffers))
    {
      *err = 0;
      return;
    }
   myGeneR[bufno]->setAccN(name);
   *err=1;
}


int GeneR_glob::getstrand() const
{
  return strand;
}

void GeneR_glob::setstrand(int _strand)
{
  strand= _strand;
}

GeneR_seq * GeneR_glob::getGeneR_seq(int bufno)
{
  return myGeneR[bufno];
}
