/*! \file readEmblLocation.cc
 *  \brief read location in an EMBL file
 *  \version 1
 *  \date Created       : 10/02/05 
 *  \date Last Modified : Time-stamp: <2007-10-11 17:09:25 antoine>
 *  \author A. Lucas
 *  \note Licence: CeCIll
 */


#include <string>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <sstream>
#include <vector>
#include "libIndex.h"
#include "GeneR_globals.h"

#include <R.h>
#include <Rdefines.h>

#define min( A , B )  ( ( A ) < ( B ) ? ( A ) : ( B ) )
#define max( A , B )  ( ( A ) > ( B ) ? ( A ) : ( B ) )


using namespace std;
/** \fn readLocation
 * \brief read Embl or Gbk location
 *
 * The same function is used for Embl or Gbk files, only difference
 * is the begining of sequence noted SQ for EMBL or ORIGIN for Gbk.
 * Look for feature "key" (by example CDS, gene, intron...) then
 * if subkey specified, extract specific subkey (i.e. locus-tag, 
 * protein, note...)
 *
 * It returns only positions delimited in from and to.
 *
 * \param file file name path
 * \param debfeature position in file of begining feature. (1st col
 *        of index file).
 * \param from, to limits for features to retrieve (integer / scalar)
 * \param key, subkey character string, key and subkey to extract
 * \param seq_origine "SQ" for embl, "OR" for genebank
 * 
 */

SEXP readLocation ( SEXP file, SEXP debFeatures, SEXP from, SEXP to,
			  SEXP key, SEXP subkey,SEXP seq_origine)

{
  SEXP ans,locis,results,strandsR,fromOut,toOut,supR,infR;

  int i,lengthkey,lengthsubkey,deb,fin,flag,strandtmp;
  string::size_type tmpentier,tmpsup,tmpinf;
  int position1, position2;
  int inkey=0;
  int fromC, toC,debFeaturesC;
  const char * fileC;
  const char * keyC;
  const char * subkeyC;
  const char * motif_seq;

  string OneLine,word,subword;

  vector<string> resultats;
  vector<string> positions;
  vector<int> strands;
  vector<int> toOutC;
  vector<int> fromOutC;
  vector<int> sup;
  vector<int> inf;


  /* Get inputs from R */
  fileC = CHAR(STRING_ELT(file,0));
  keyC = CHAR(STRING_ELT(key,0));
  subkeyC = CHAR(STRING_ELT(subkey,0));
  motif_seq = CHAR(STRING_ELT(seq_origine,0));
 
  fromC = INTEGER(from)[0];
  toC = INTEGER(to)[0];
  debFeaturesC = INTEGER(debFeatures)[0]; /* Must be first column of index file */
  
  lengthkey = strlen(keyC);
  lengthsubkey = strlen(subkeyC);

  
  ifstream thefichier (fileC,ios::in);
  if(!thefichier)
    {
      cout << "GeneR.so: open file error";
      return(libIndex::returnInteger(0));
    }

  thefichier.seekg(debFeaturesC,ios::beg);
  if(!thefichier)
    {
      cout << "GeneR.so: seek in file error";
      thefichier.close();
      return(libIndex::returnInteger(0));
    }
  getline(thefichier,OneLine,'\n');
  

  while(thefichier)
    {
      if( OneLine.find(motif_seq) ==0)
	  break;


      if(((int)OneLine.find(keyC))==5)
	{
	  /* we found key */
	  inkey=1;
	  position1 = OneLine.find("join");
	  position2 = OneLine.find("complement");
	  if(!(position1 < (int) OneLine.size()))
	     position1 = -1;
	  if(!(position2 < (int) OneLine.size()))
	     position2 = -1;
	  if( ( position1 >= 21) || (position2 >=21) )
	    {
	      /* Penible cas: position is complicated */
	      /* look for begin */
	      
	      tmpentier = max(position1+5,position2+11);
	      if(position2>=21)
		  strandtmp = 1;
	      else
		  strandtmp = 0;

	      word  = OneLine.substr(tmpentier,OneLine.size());

	      istringstream tampon (word);
	      if(tampon >> deb)
		tmpinf = 0;
	      else
		{
		  tmpinf = 1;
		  istringstream tampon (OneLine.substr(tmpentier+1,OneLine.size()) );
		  tampon >> deb;
		}

	      /* look for end */ 
	      while((int) word.find(")") == -1 )
		{
		  if(! thefichier)
		    break;
		  getline(thefichier,OneLine,'\n');
		  word = word + OneLine.substr(21,OneLine.size());
		}
	    }
	  else
	    {
	      /* case: just one position */
	      word  = OneLine.substr(21,OneLine.size());
	      istringstream tampon (word);
	      if(tampon >> deb)
		tmpinf = 0;
	      else
		{
		  tmpinf = 1;
		  istringstream tampon (OneLine.substr(22,OneLine.size()) );
		  tampon >> deb;
		}
	      strandtmp=0;
	    }
	  i = max( ((int)word.rfind(".")) , ((int)word.rfind(",") )); 

	  tmpsup = 0;
	  tmpentier = word.find(">");
	  if(tmpentier >=0 && tmpentier <= word.size())
	    tmpsup = 1;

	  subword = word.substr(i+1+tmpsup,OneLine.size()-i);
	  
	  istringstream tampon2 (subword);
	  tampon2 >> fin;
	  
	  if((deb >= fromC) && (fin <= toC))
	    {
	      word = word.substr(0,word.find(")"));	      
	      positions.push_back(word);
	      strands.push_back(strandtmp);
	      fromOutC.push_back(deb);
	      toOutC.push_back(fin);	      
	      inf.push_back(tmpinf);
	      sup.push_back(tmpsup);
	    }
	  else
	    inkey=0;

	  /* look for the subkey */
	  if((lengthsubkey>0) && inkey)
	    {
	      while(thefichier)
		{
		    getline(thefichier,OneLine,'\n');
		    if(OneLine.size() < 7)
		    {
		      resultats.push_back("");
		      break;
		    }

		    if((OneLine.substr(5,3) != "   ")||( OneLine.find(motif_seq)==0))
		    {
		      thefichier.seekg(-(OneLine.size()+1),ios::cur);
		      resultats.push_back("");
		      break;
		    }

		  /* something like \subkey="bla bla bla" */
		  if(OneLine.find(subkeyC)== 22)
		    {		      
		      position1 = OneLine.find("\"") ;
		      if((position1 > 0 ) && (position1 <= (int) OneLine.size() ))
			{
			  word= OneLine.substr(23+lengthsubkey,OneLine.size());
			  if(((int)OneLine.find("\"")) ==((int) OneLine.rfind("\"")))
			    {
			      flag = 1;
			      position1=OneLine.find("\"");
			      while( (position1 < 0) || (position1 > (int) OneLine.size())  || flag)
				{
				  flag = 0;
				  if(! thefichier)
				      break;
				  if((OneLine.substr(5,3) != "   ")||( OneLine.find(motif_seq)==0))
				    {
				      
				      thefichier.seekg(-OneLine.size(),ios::cur);
				      cout << "GeneR.so: warning\n";
				      break;
				    }
				  getline(thefichier,OneLine,'\n');
				  position1=OneLine.find("\"");
				  word =word+OneLine.substr(21,OneLine.size());
				}
			    }
			  /* delete guillements */
			  word = word.substr(1,word.size()-2);
			}
		      else		
			  word = OneLine.substr(23+lengthsubkey,OneLine.size());

		      resultats.push_back(word);
		      break;
		    }
		}
	    }
	}

      if(thefichier)
	getline(thefichier,OneLine,'\n');

      
    }

  thefichier.close();

  if(positions.size()==0)
    return(libIndex::returnInteger(0));

  

  PROTECT(locis = allocVector(STRSXP, positions.size()));
  PROTECT(strandsR = Rf_allocVector(INTSXP,positions.size()));
  PROTECT(fromOut= Rf_allocVector(INTSXP,positions.size()));
  PROTECT(toOut= Rf_allocVector(INTSXP,positions.size()));
  PROTECT(infR= Rf_allocVector(INTSXP,positions.size()));
  PROTECT(supR= Rf_allocVector(INTSXP,positions.size()));
  for(i =0; i< (int)positions.size(); i++)
    {
      SET_STRING_ELT(locis, i, mkChar( positions[i].c_str() ));
      INTEGER(strandsR)[i] = strands[i];
      INTEGER(fromOut)[i] = fromOutC[i];
      INTEGER(toOut)[i] = toOutC[i];
      INTEGER(supR)[i] = sup[i];
      INTEGER(infR)[i] = inf[i];
    }
  UNPROTECT(6);

  /* free vectors */
  positions.clear();
  strands.clear();
  fromOutC.clear();
  toOutC.clear();
  sup.clear();
  inf.clear();
  /* free strings */
  /* OneLine.clear();
  word.clear();
  subword.clear();*/

  if(!lengthsubkey)
    {
      resultats.clear();
      PROTECT(ans = NEW_LIST(6));
      SET_VECTOR_ELT(ans, 0,locis);
      SET_VECTOR_ELT(ans, 1,strandsR);
      SET_VECTOR_ELT(ans, 2,fromOut);
      SET_VECTOR_ELT(ans, 3,toOut);
      SET_VECTOR_ELT(ans, 4,infR);
      SET_VECTOR_ELT(ans, 5,supR);
      UNPROTECT(1);
      return(ans);
    }


  PROTECT(results = allocVector(STRSXP, resultats.size()));
  for(i =0; i< (int) resultats.size(); i++)
    SET_STRING_ELT(results, i, mkChar( resultats[i].c_str() ));
  UNPROTECT(1);

  resultats.clear();
  
  PROTECT(ans = NEW_LIST(7));
  SET_VECTOR_ELT(ans, 0,locis);
  SET_VECTOR_ELT(ans, 1,results);
  SET_VECTOR_ELT(ans, 2,strandsR);
  SET_VECTOR_ELT(ans, 3,fromOut);
  SET_VECTOR_ELT(ans, 4,toOut);
  SET_VECTOR_ELT(ans, 5,infR);
  SET_VECTOR_ELT(ans, 6,supR);
  UNPROTECT(1);

  return(ans);

}
