/*! \file readSeqEmbl.cc
 *  \brief functions to read Embl file
 *
 *  \version 1
 *  \date Created       : 26/10/05 
 *  \date Last Modified : Time-stamp: <2007-10-12 18:38:44 antoine>
 *  \author A. Lucas from "Bio-R" codes
 *  \note Licence: CeCIll
 */

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <ctype.h>


#include "GeneR_glob.h"
#include "readSeqEmbl.h"


#include "libStrings.h"
#include "libIndex.h"
#include "GeneR_globals.h"

extern "C"
{
  void ReadSeqEMBLR(char **file,int *offsq, int *begseq, int *endseq, 
		    int *bufno, int * upper, int *err);

  void ReadSeqEMBL(char ** seq,char ** BqueFich, int *OffsSQ, 
		       int *DebSeq, int *FinSeq, int* upper, int *err);


}

/* ---------------------------------------------------- */

/* ---------------------------------------------------- */
/*                      ReadSeqEMBLR                    */
/* ---------------------------------------------------- */
/** @brief Read Embl
 *
 * This function allocate a buffer 
 * and call ReadSeqEMBL. This function
 * is called directly by R.
 *  
 *   @param **file Embl file
 *   @param *offsq    Offset sequence (i.e. position of sequence
 *                    first char in file)
 *   @param *begseq   position to begin in sequence
 *   @param *endseq   position to end in sequence
 *   @param *bufno    Integer, number of the sequence. (bufseq)
 *   @param *upper    Flag 1: all character are converted to upper case
 *   @param *err      Error Flag: to return to R.
 *
 *  \sa ReadSeqEmbl
 */
/* ---------------------------------------------------- */
void ReadSeqEMBLR(char **file,int *offsq, int *begseq, 
		  int *endseq, int *bufno, int * upper, int *err)
{
  int l;
  char * seq;

  /* size to alloc = size of seq + 1 = end - beg +2*/
  l = *endseq-*begseq+2;
  

  if(! GeneR_glob::instance()->allocBuffer(l, *bufno, 0, 1))
    {
      *err= 0;
      printf("GeneR: Error in function ReadSeqEMBLR\n");
      return;
    }
  /*      seq = &(sBuffer[*bufno]);*/
  seq = GeneR_glob::instance()->buffer(*bufno);

  /*  printf("param %d %d %d \n",*offsq,*begseq,*endseq);*/
      
  ReadSeqEMBL(&seq, file,offsq,begseq,endseq,upper,err);
      
  /*      sSize [*bufno] = *endseq - *begseq + 1;*/
      
  return ;

}



/** @brief Acces direct a un extrait de sequence dans 
    une banque. Installe la sequence dans le buffer 0.
	   Si  Argument FinSeq == 0, lire toute la sequence 
           Retourne 0 si OK, -1 si pas de sequence extraite 
	   @param *seq   Address of sequence
    @param *BqueFich Path du fichier banque
    @param *OffsSQ   Position dans le fichier du debut de
                          la sequence
    @param *DebSeq  Debut de l'extraction
    @param *FinSeq  Fin de l'extraction
    @param *upper Flag 1: all character are converted to upper case
    @param *err Flag to return (0 if Ok, -1 if not)
*/
/* ----------------------------------------------------- */
void ReadSeqEMBL(char ** seq,char ** BqueFich, int *OffsSQ, 
	      int *DebSeq, int *FinSeq, int * upper, int *err)
{
  FILE *f;			/* fichier courant */
  char * LBuf;		/* Ligne courante */
  char sBuf[MaxLLigne];		/* tampon pour sequence */
  char tBuf[MaxLLigne];		/* locale */
  int LLBuf, LsBuf;		/* Taille de la ligne courante & de sBuf */
  int PtDeb, PtFin;		/* Pour extraire les nt d'une ligne */
  int NumntFinL;	        /* numero de nt debut & fin de ligne courante */
  int FinEntree, Continue;	/* Drapeaux de fin d'entree & contineur l'extraction */
  int longEnregistrement=0;       /* Compte la longeur de l'enregistrememnt (verifie */
                                /* l'intgrite */
  int tailleSeq=0;              /* Size of all sequence */


  LBuf = (char*) malloc(MaxLLigne * sizeof(char));
/* Ouvrir la banque et pointer le debut de sequence */


  

  if ( !(f = fopen(*BqueFich, "r")))
    {
      printf("GeneR.so: error while opening file\n");
      *err = -1;
      return;
    }
	
#ifndef __MINGW_H
  if ( (fseek(f, *OffsSQ, 0))) 
    {
      printf("GeneR.so: error while seeking into file\n");
      fclose(f);
      *err = -1;
      return ;
    }
#else
  /* fseek on windows count CR ('\r') and fgets does not see them */
  int i;
  for(i=0; i< *OffsSQ; i++)
    if(fgetc(f) == EOF)
      {
      printf("GeneR.so: error while seeking file\n");
      fclose(f);
      *err = -1;
      return ;
      }
#endif
   
  /* Prendre la ligne courante (verifier que pas "//") */
  
  LBuf = fgets(LBuf,MaxLLigne,f);
  if (readSeqEmbl::EstFinEntreeEMBL(LBuf))
    {
      fclose(f);
      *err = -1;
      return ;	/* pas de sequence */
    }
  LLBuf = strlen(LBuf);

  /* Extraire le numero de nucleotide qui se trouve en bout de ligne */
  
  if ((NumntFinL = readSeqEmbl::XtNumntFinLig(LBuf,1))==-1) 
    {
      fclose(f);
      *err = -1;
      return ;
    }
  
  /* Lire ligne suivante tant que le n de nt de fin de ligne < DebSeq */

  while (NumntFinL < *DebSeq) {
    
    LBuf = fgets(LBuf,MaxLLigne,f);

    if (readSeqEmbl::EstFinEntreeEMBL(LBuf))
      {
	fclose(f);      
	*err = -1;
	return ; /* pas de sequence */
      }
    if ((NumntFinL = readSeqEmbl::XtNumntFinLig(LBuf,1)) == -1) 
      {
	fclose(f);
	*err = -1;
	return ;
      }
  } 



  /* XtNumntFinLig Extrait le numero du dernier nt de la ligne (format
     EMBL) et raccourcit la ligne de ce numero*/
  
  /* Ici, la ligne courante contient le nt de DebSeq
     Compacter les nt de la ligne, extraire les nt utiles dans sBuf */ 
  if(NumntFinL<0)
    { 
      fclose(f);
      *err = -1;
      return  ;
    }
  
  if (!(LsBuf = libStrings::SupprimerSeparateurs(LBuf, sBuf)))
    {
      fclose(f);
      *err = -1;
      return ;
    }


  
  PtDeb = LsBuf - NumntFinL + *DebSeq - 1;
  
  if ((*FinSeq > 0) && (*FinSeq < NumntFinL)) 
    PtFin = LsBuf - NumntFinL + *FinSeq - 1;
  else 
    PtFin = LsBuf - 1;  
  
  /*  readSeqEmbl::strxtr(sBuf, tBuf, PtDeb, PtFin -PtDeb + 1);*/
  strncpy(tBuf,sBuf+PtDeb,PtFin -PtDeb + 1);
  tBuf [PtFin -PtDeb + 1] = '\0';

  tailleSeq = (*FinSeq)- (*DebSeq) +1;
  longEnregistrement =   strlen(tBuf);
  if(longEnregistrement> tailleSeq)
    {
      printf("Not enough allocation. [Probably: Verify type of file (fasta, emlb...) and delete index file '.ix']\n");
      fclose(f);
      *err = -1;
      return ;
    }



  /*PousseMot(tBuf,*seq); */  		    /* pour ajouter au buffer 0 */
  *seq = strcpy(*seq,tBuf);  
  /* Recommence jusqu'a la fin de l'entree */
  
  Continue = ((NumntFinL < *FinSeq) || (*FinSeq == 0));
  

  while (Continue) {
    LBuf = fgets(LBuf,MaxLLigne,f);
    
    FinEntree = (readSeqEmbl::EstFinEntreeEMBL(LBuf)); 
    if (!FinEntree) {
      longEnregistrement = libStrings::Pousse_atgc(longEnregistrement,LBuf,*seq,tailleSeq);
      
    }
    Continue = !FinEntree && (longEnregistrement < tailleSeq); 
  } 

  fclose(f);
  free(LBuf);
  /*DEBUG */
  /*  printf("fichier ferme\n"); */
  
  if(*upper)
    libStrings::sys_upper_string(*seq);

  *err = 1;
  return ;

} 
namespace readSeqEmbl
{

  
  /* ---------------------------------------------------- */
  /*                     XtNumntFinLig	                */
  /* ---------------------------------------------------- */
  /** @brief Extrait le numero du dernier nt de la ligne
      (format EMBL) et raccourcit la ligne de ce numero
      + Idem si le numero est au debut.
      Renvoie -2 si la ligne fait moins que 3 charactres
      (permet de sauter les lignes vides)
      Renvoie -2 si le 2eme charactere de la ligne n'est pas 
      un sparateur. (permet de sauter des lignes de commentaires)
      Renvoie -1 s'il y a une erreur.
      @param  *s Ligne de sequence Embl
      @param  getnumber: 1 return number, 0, just troncate line (to
      go more fast...)
  */
  int XtNumntFinLig(char *s,int getnumber)
  {
    int n;
    int i, j, k, ls,lt;
    char t[MaxLNom];
    
    i = strlen(s) - 1;
    
    if(i < 4)
      return -2;

    if(isalnum(s[1]) )
      return(-2);
    
    while ((!isalnum(s[i])) && (i>0))  s[i--] = '\0';

    if(isdigit(s[i]))
      {
	/* Le numero est  la fin de la ligne */
	while(isdigit(s[i]) && (i>0))  i--;
      
	if(getnumber)
	  {
	    k = 0;
	    ls = strlen(s);
	    for(j = i; j < ls; j++) t[k++] = s[j];  
	    t[k] = '\0';	/* termine t */
	    s[i] = '\0';	/* raccourcit s */
	    if (sscanf(t, "%d", &n))
	      return n;
	    else return -1;
	  }
	else {
	  s[i] = '\0';	/* raccourcit s */
	  return(1);
	}
      }
    else
      {
	/* On regarde si le numero est au debut de la ligne */
	i = 0 ;
	ls = strlen(s) ;
	k = 0;
	lt=0;
	while((!isalnum(s[i])) && (i < ls) )  i++;
	
	/* s'il n'y a pas de chiffres ni au debut, ni a la fin */ 
	if(!isdigit(s[i+lt]))
	  return(-2);
	while(isdigit(s[i+lt]) && ((i+lt) <= ls) )  lt++;
	/* On copie dans t le numero */
	for(j = 0; j < lt; j++) t[j] = s[j+i];  
	t[lt]= '\0';
	
	/* On decale la ligne s (on ecrase les chiffres) */
	ls = strlen(s);
	k = 0;
	for(j = 0; j <= ls-i-lt; j++)
	  {
	    s[j] = s[j+i+lt];  
	    /* On compte le nb de separateurs (dans k)*/
	    if(!isalnum(s[j]))
	      k++;
	  }
      
	/* numero du dernier nt = numero du debut + longeur de 
	   l'ensemble - nb de nombres - nb de separateurs*/ 
	if (sscanf(t, "%d", &n)) 
	  {
	    /*	  printf("nb %d\n",n+ls-i-k-lt-1);*/
	    return (n+ls-i-k-lt-1);
	  }
	else return -1;
      }
  }  

  /* --------------------------------------------------- */
  /*                 EstFinEntreeEMBL	               */
  /* --------------------------------------------------- */
  /** @brief vrai si "s" commence par "//" ou si fin de 
      fichier
      @param  *s Ligne a tester
  */
  int EstFinEntreeEMBL(char *s)
  {
    if((s == NULL) || (s ==  (char *) EOF)) 
      return 1; /* Fin de Fichier */ 
    if (strlen(s) < 2) 
      return 0;
    else if ((s[0] == '/') && (s[1] == '/')) return 1;
    else return 0;
  }  
  

  /* ---------------------------------------------------- */
  /* @Function: strxtr                                
   peux etre evite ! : utilisation de strncpy(dest,src+deb,taille)
   puis ajout d'un '\0' a la fin*/

}/* End of Namspace */







