/*! \file readSeqFasta.cc
 *  \brief  Some functions for Fasta file
 *
 *  \date Created       : 26/10/05 
 *  \date Last Modified : Time-stamp: <2007-10-12 18:39:49 antoine>
 *  \author A. Lucas from "Bio-R" codes
 *  \note Licence: CeCIll
 */


#include <stdio.h>
#include <string>
#include "GeneR_globals.h"
#include "readSeqFasta.h"
#include "GeneR_glob.h"
#include "libStrings.h"
#include <R.h>
#include <Rdefines.h>

extern "C"
{
  void ReadSeqFASTAR (char **file,int *offsq,
		      int *begseq, int *endseq, int *bufno, 
		      int * upper, int *err);

  void ReadSeqFASTA(char ** seq, char **path,int *offsetseq, int *begseq,
		    int *endseq, int * upper, int *err);
  
  SEXP fasta_descript (SEXP file,SEXP label, SEXP deb, SEXP fin) ;

}



/* ---------------------------------------------------- */
/*                   ReadSeqFASTAR                      */
/* ---------------------------------------------------- */
/** @brief Read Fasta. 
 *
 * This function allocate a buffer 
 * and call ReadSeqFASTA. This function
 * is called directly by R.
 *
 * \param  **file  fasta file
 * \param  *offsq   Offset sequence (i.e. position of sequence
 *                  first char in file)
 * \param  *begseq  position to begin in sequence
 * @param  *endseq  position to end in sequence
 * @param *bufno    Integer, number of the sequence. (bufseq)
 * @param *upper    Flag 1: all character are converted to upper case
 * @param *err      Error Flag: to return to R.
 */
/* ---------------------------------------------------- */

void ReadSeqFASTAR (char **file,int *offsq,
		    int *begseq, int *endseq, int *bufno, int * upper, int *err)
{
  char *seq;

  if(!GeneR_glob::instance()->allocBuffer((*endseq)-(*begseq) + 2, *bufno, 0,1))
    {
      *err = -1;
      return;
    }

  seq = GeneR_glob::instance()->buffer(*bufno,0);

  ReadSeqFASTA (&seq,file,offsq,begseq, endseq,upper,err);
  
  if(seq) {
    *err=1;

    /*if(*upper == 1)
      sys_upper_string(sBuffer[*bufno]);*/
  }
  else 
    {
      GeneR_glob::instance()->freeSeq(*bufno);
      *err=-1;
    }  

  return;

}


/* ------------------------------------------------------ */
/* @Function: ReadSeqFASTA	                          */
/** @brief  Acces direct a un extrait de sequence dans  
   une banque. Installe la sequence dans le buffer 0 
   Si  Argument FinSeq == 0, lire toute la sequence  
   Procedure de pilotage et gestion du buffer        */
/* ------------------------------------------------------ */

void ReadSeqFASTA(char ** seq, char **path,int *offsetseq, int *begseq,
		   int *endseq,  int *upper, int* err)
{
  FILE *stream;
  char *buffer;  
  char *bigBuffer;
  int lenBigBuffer, lenBuffer, length, pos, from, to;
  
  from=*begseq - 1;
  to = *endseq - 1;
  pos = -1;
  length = *endseq - * begseq + 1;

  if (!(stream = fopen(*path, "r")))
    {
      fprintf(stdout, "GeneR.so: error while opening file\n");
      *err = -1;
      *seq = NULL;
      return ;
    }


#ifndef __MINGW_H
  if (fseek(stream, *offsetseq, SEEK_SET))
    {
      fprintf(stdout, "GeneR.so: error seeking into file\n");
      *err = -1;     
      *seq = NULL;
      fclose(stream);
      return ;
    }
#else
  int i;
  for(i=0; i< *offsetseq; i++)
    if(fgetc(stream) == EOF)
    {
      fprintf(stdout, "GeneR.so: error seeking into file\n");
      *err = -1;     
      *seq = NULL;
      fclose(stream);
      return ;
    }
#endif
  /*
  if(!(bigBuffer = (char *)malloc(length * sizeof(char) + 1)))
    {
      fprintf(stdout, "GeneR.so: memory allocation error in ReadSeqFASTA\n");
      *bufno = -1;
      *seq = NULL;
      return ;
    }
  */
  bigBuffer = *seq;

  lenBigBuffer = 0;
  
  while ((buffer = readSeqFasta::sNextIOBuffer(stream, 0, 0)) && 
	 (pos < (*endseq))) 
    {      
      if (*buffer == '>') {
	(void) readSeqFasta::sNextIOBuffer(stream, 1, 1); /* push it back */
	break;
      }
      
      lenBuffer = strlen(buffer);
      
      /*printf("%s\n", buffer);*/

      pos = pos + lenBuffer;
      /*printf("lenBuffer = %d\tpos = %d\tfrom=%d\tto=%d\n", */
       /*   lenBuffer, pos, from, to);*/

      if(pos>=from) {
	if(from >= pos - lenBuffer + 1) {
	  if(pos<to) {
	    /*printf("pos < to, lenBuffer = %d\tpos = %d\tfrom=%d\tto=%d\n", */
	    /* lenBuffer, pos, from, to);*/
	    (void) memcpy(bigBuffer + lenBigBuffer, 
			  buffer+(lenBuffer - (pos-from+1)), pos-from+1);
	  
	    lenBigBuffer = lenBigBuffer + (pos-from)+1;
	    bigBuffer[lenBigBuffer] = '\000';
	    /*printf("%s\n", buffer);*/
	  }
	  else {
	    (void) memcpy(bigBuffer + lenBigBuffer, 
			  buffer+(lenBuffer - (pos-from+1)),
			  to-from+1);
	    lenBigBuffer = lenBigBuffer + to-from+1;
	    bigBuffer[lenBigBuffer] = '\000';
	    /*printf("%s\n", buffer);
	      printf("bigBuffer : %s\n", bigBuffer); */
	  }
	}
	else {
	 if(pos<to) {
	    (void) memcpy(bigBuffer + lenBigBuffer, 
			  buffer,lenBuffer);
	  
	    lenBigBuffer = lenBigBuffer + lenBuffer;
	    bigBuffer[lenBigBuffer] = '\000';
	    /*printf("%s\n", buffer);*/
	  }
	  else {
	    (void) memcpy(bigBuffer + lenBigBuffer, 
			  buffer, lenBuffer-(pos-to));
	    lenBigBuffer = lenBigBuffer +lenBuffer-(pos-to);
	    bigBuffer[lenBigBuffer] = '\000';
	    /*printf("%s\n", buffer);*/
	  } 
	}
      }
    }


  bigBuffer[lenBigBuffer] = '\000';

  fclose(stream);
  
  *seq = bigBuffer;

  if(*upper == 1)
    libStrings::sys_upper_string(*seq);

  *err= 1;

  return;
  
  }



namespace readSeqFasta{
  
  static int sRetained = 0;

  /** @brief lecture bufferisee                   */
  /* -------------------------------------------- */
  char * sNextIOBuffer(FILE *streamin, int retain, int serial)
  {
    char   *buf, *end;
    int   reused;
    static char sBuffer[BUFSIZ];  
    
    reused = (retain || sRetained) && serial;
    
    sRetained = retain;
    
    buf = (   reused 
	      ? sBuffer 
	      : fgets(sBuffer, sizeof(sBuffer), streamin));	 	
    
    if (! buf)		  /* end of file	    */
      return NULL;
    
    end = buf + strlen(buf) - 1;
    
    if (*end == '\n')	  /* remove trailing \n	    */
      *end = '\000';

    return (  *buf || reused
	      ? buf
	      : sNextIOBuffer(streamin, retain, serial));
  }
}



/** @brief Get Fasta description (of a sequence)
 *
 *   @param file, label: Character string
 *   @param deb, fin: integers
 *   \return  (ans) String return -1 if error description if not.
 */
SEXP fasta_descript (SEXP file,SEXP label, SEXP deb, SEXP fin) 
{
  SEXP ans;
  char * mot;

  FILE *fichier;
  int longueur,l1,l2,len,i;


  PROTECT (file = AS_CHARACTER(file));
  PROTECT (label = AS_CHARACTER(label));
  PROTECT (deb = AS_NUMERIC(deb));
  PROTECT (fin = AS_NUMERIC(fin));

  len = strlen( CHAR(STRING_ELT(label,0)));
  
  l1 =(int) NUMERIC_POINTER(deb)[0];
  l2 =(int) NUMERIC_POINTER(fin)[0];
  fichier=fopen(CHAR(STRING_ELT(file,0)),"r");

#ifndef __MINGW_H
  fseek(fichier, l1+len+2, SEEK_SET);
#else
  for(i=0; i< l1+len+2; i++)
    if(fgetc(fichier) == EOF)
#endif

  UNPROTECT(4);
  
  /* l2 = 1er character de la sequence=> l2-2 = dernier character du label (\n)
   * debut: l1+len+2 (l1 = position de '>' que l'on passe + label + espace
   * total de character longueur = fin - debut +1  = (l2-2) - (l1+len+2) +1
   * allocation long+1 (\0) en plus
   */
  longueur = l2-l1-len-3  ;
  if(longueur >0)
    {
      mot = (char *) malloc( (longueur+1) * sizeof(char));
      if(mot == NULL)
	{
	  printf("GeneR: Memory allocation error at fasta_descript\n");
	  return (0);
	}


      for(i=0;i<longueur;i++)
	{
	  mot[i] = (char) fgetc(fichier);	  
	}      
      mot[longueur]='\0';
    }
  else
    {
      mot = (char *) malloc( 1 * sizeof(char));
      mot[0] = '\0';
    }
  fclose(fichier);

  PROTECT (ans = allocVector(STRSXP,1));
  SET_STRING_ELT (ans,0,mkChar(mot));  
  UNPROTECT(1);

  free(mot);
  return(ans);


}
