/*! \file seqManip.cc
 *  \brief functions to maniplate sequence buffers
 *
 *  \version 1
 *  \date Created       : 26/10/05 
 *  \date Last Modified : Time-stamp: <2007-10-12 18:41:32 antoine>
 *  \author A. Lucas from "Bio-R" codes
 *  \note Licence: CeCIll
 */

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <R.h>
#include <Rinternals.h>          /* headers pour les fonctions de R */
#include <Rdefines.h>

#include "GeneR_glob.h"
#include "seqManip.h"
#include "libStrings.h"

/*
#include "libIndex.h"
#include "GeneR_globals.h"*/

extern "C"
{
  SEXP multiextract (SEXP bufno,SEXP vfrom,SEXP vto, SEXP comp);
  void assemble (int *bufseq, int *from, int *to, int *bufdest, int *strand,int *err);
  void sys_concat(int *seq_one, int *seq_two, int *seq_dest, 
	      int *from1, int *to1,
	      int *strand1, int *from2, int *to2, int *strand2);
  void concat(int * seqdestno, int * seqno, int *error);
  void sys_placestring (char **s, int *bufno,int * upper);
}


/** @brief extrait des sous sequences en fonction des
           vecteurs vfrom et vto et les place dans un
	   vecteur de chaine de caracteres
	   < LC > add 15 Feb 00
    @param bufno  Numero de la sequence
    @param vfrom Vecteur de debuts de l'extraction
    @param vto   Vecteur de fins de l'extraction
    @param comp  0 :  brin sens - 1 : brin anti-sens
*/
/* ---------------------------------------------------- */

SEXP multiextract (SEXP bufno,SEXP vfrom,SEXP vto, SEXP comp)
{
  int i = 0;
  char * seq;
  char *substr;
  SEXP strings;
  SEXP err;
  int n, seqno,strand,maxlong;
  int *from, *to;

  PROTECT(err = NEW_INTEGER(1));
  PROTECT(bufno = AS_INTEGER(bufno));
  PROTECT(vfrom = AS_INTEGER(vfrom));
  PROTECT(vto = AS_INTEGER(vto));
  PROTECT(comp = AS_INTEGER(comp));
  
  n = LENGTH(vfrom);

  PROTECT(strings = NEW_CHARACTER(n));

  seqno = INTEGER(bufno)[0] ;
  strand = INTEGER(comp)[0];


  seq = GeneR_glob::instance()->buffer(seqno,strand);


  if (seq == NULL)
    {
      INTEGER(err)[0] = -1;     /* valeur d'erreur dans R */
      UNPROTECT(6);
      return (err);
    }

  from = INTEGER_POINTER(vfrom);
  to = INTEGER_POINTER(vto);
 
  maxlong = 0; /* Length of larger sequence; maxlong +1: "\0" */
  for(i = 0 ; i<n; i++)
    if(from[i] <=  to[i] && from[i] > 0)
      if(to[i]-from[i]+1 > maxlong)
	maxlong = to[i]-from[i]+1 ;

  substr = (char *)malloc((maxlong + 1)*sizeof(char));
  i=0;

  while (i < n)
    {
      if(from[i] <=  to[i] && from[i] > 0 && to[i] < GeneR_glob::instance()->size(seqno,strand))
	{
	  strncpy(substr,seq+from[i]-1,to[i]-from[i]+1);
	  substr[to[i]-from[i]+1] ='\0';
	  SET_STRING_ELT(strings, i, mkChar(substr));
	}
      else
	SET_STRING_ELT(strings, i, mkChar(""));     
      i++;
    }
  
  free(substr);
  UNPROTECT(6);
 
  return(strings);


}


/* ------------------------------------------------------------------------ */ 
/*                              assemble		                    */
/* ------------------------------------------------------------------------ */ 
/** @brief cre une nouvelle squence en concatnant les rgions
           prcises par from et to de la squence contenue dans bufseq 
	   [LC] created 25 Oct 00
    @param  *bufseq    Numero de la sequence d'origine
    @param  *from      Positions de debut des regions a assembler
    @param  *to        Positions de fin des regions a assembler
    @param  *bufdest   Numero du bufseq de destination (deja alloue)
    @param  *strand    0 : brin sens. 1 : brin anti-sens
    @param  *err       0:error;  1: good
*/
/* ------------------------------------------------------------------------ */

void assemble (int *bufseq, int *from, int *to, int *bufdest, int *strand,int *err)
{


  char * seqdest;
  char * seqsource;

  int size;

  /*  tmpseq= (char * ) malloc((*to - (*from) +3)*sizeof(char));*/
  
  seqdest = GeneR_glob::instance()->buffer(*bufdest, 0);
  seqsource = GeneR_glob::instance()->buffer(*bufseq, *strand);
  
  if((seqdest == NULL) || (seqsource == NULL))
    {
      *err=0;
      return;
    }
  size = strlen( seqdest);
  strncpy(seqdest+size,seqsource+(*from)-1,(*to)-(*from)+1);

  seqdest[size+(*to) - (*from) +1] = '\0';

  *err = 1;

  return;
}

/** @brief Concatene 2 sequences dans une troisieme	
           < LC > created Dec 00
    @param  *seq_one  Numero de la premiere sequence
    @param  *seq_two  Numero de la deuxieme sequence
    @param  *seq_dest Numero de la sequence de destination
    @param  *from1    Positions de debut sur la premiere sequence
    @param  *from2   Positions de debut sur la deuxieme sequence
    @param  *to1      Positions de fin sur la premiere sequence
    @param  *to2      Positions de fin sur la deuxieme sequence
    @param  *strand1  Indication du brin pour chaque region de la 
	              premiere sequence
	   		 0 : brin sens. 1 : brin anti-sens
    @param  *strand2  Indication du brin pour chaque region de la 
	                 deuxieme sequence
			 0 : brin sens. 1 : brin anti-sens
*/
/* --------------------------------------------------------------------	*/

void sys_concat(int *seq_one, int *seq_two, int *seq_dest, 
		int *from1, int *to1,
		int *strand1, int *from2, int *to2, int *strand2)
{


  char *seq1;
  char *seq2;
  char *seq3;
  int longueur;



 seq1 = GeneR_glob::instance()->buffer(*seq_one,*strand1) + *from1 -1;
 seq2 = GeneR_glob::instance()->buffer(*seq_two,*strand2) + *from2 -1;

 
 longueur = *to1 - *from1 + *to2 - *from2 + 3;

 if(! GeneR_glob::instance()->allocBuffer(longueur, *seq_dest, 0, 1))
   return;

 seq3 = GeneR_glob::instance()->buffer(*seq_dest,0); 

 seq3 = strncpy(seq3, seq1, *to1 - *from1 + 1);
 seq3 [*to1 - *from1 + 1] = '\0';
 seq3 = strncat(seq3, seq2, *to2 - *from2 + 1);
			
 return;
}


/** Concat concat two sequences
 * \brief append buffer seqno to buffer seqdestno
 * \param seqdestno,seqno integer, number of buffer
 * \param error: flag error -1 or -2 (memory error); 0= good
 */

void concat(int * seqdestno, int * seqno, int *error)

{
  char * seqdest;
  char * seq;
  int long_source,long_dest,long_tot;


    
  /* free the rev. complement */
  GeneR_glob::instance()->freeCompSeq(*seqno);

  /* Now working... */
  seqdest  = GeneR_glob::instance()->buffer(*seqdestno);
  

  long_dest = GeneR_glob::instance()->size(*seqdestno) ;
  long_source = GeneR_glob::instance()->size(*seqno) ;
  /* note: sSize[*seqno] = strlen(seq) 
   * seqdest[long_dest] = '\0'
   */

  long_tot = long_source + long_dest-1;

  if(!GeneR_glob::instance()->allocBuffer(long_tot,*seqdestno, 0, 0))
    {
      *error = -2;
      return;
    }
     
  seq  = GeneR_glob::instance()->buffer(*seqno);
  seqdest  = GeneR_glob::instance()->buffer(*seqdestno);


  if(seqdest==NULL)
    {
      printf("Memory allocation error in GeneR.so\n");
      *error = -2;
      return;
    }

  memcpy(seqdest+long_dest-1,seq , long_source );

  *error = 0;

   /*   printf(">%s<\n",seqdest);*/

}


/* ---------------------------------------------------- */
/*              sys_placestring				*/
/* ---------------------------------------------------- */
/** @brief Place une sequence dans un bufseq
           < LC > add Feb 00
    @param **s   Sequence a placer dans le bufseq
    @param *bufno Numero du bufseq dans lequel
	              on veut placer la sequence
    @param *upper Flag 1: all character are converted to upper case
*/
/* ---------------------------------------------------- */
void sys_placestring (char **s, int *bufno,int * upper)
{

  int size = strlen(*s)+1;
  int comp;
  char * seq;
	
  comp = 0;

  if ( ! *s)
    {
      *bufno = -1;
      return;
    }
	
  
  if (!GeneR_glob::instance()->allocBuffer(size,*bufno, 0, 1))
    {
      printf("GeneR.so allocation error\n");
      *bufno =-1;
      return;
    }
  seq =  GeneR_glob::instance()->buffer(*bufno,0);


  strcpy (seq, *s);
	
  if(*upper)
    libStrings::sys_upper_string(seq);
	
}

