/*! \file vecteurs.cc
 *  \brief functions to manipulate segments sets.
 *  by lower / upper case.
 *  \version 1
 *  \date Created       : 26/10/05 
 *  \date Last Modified : Time-stamp: <2006-03-03 17:55:55 lucas>
 *  \author  A. Lucas
 *  \note Licence: CeCIll
 */

#include <stdio.h>

#define max( A , B )  ( ( A ) > ( B ) ? ( A ) : ( B ) )
#define min( A , B )  ( ( A ) < ( B ) ? ( A ) : ( B ) )

extern "C"
{
  void vec_union (double * deb, double * fin, int * n,int * flag);
  void vec_minimal (double * deb, double * fin,  int * n);
  void relist (int *  envdeb, int * envfin, int * posdeb, int * posfin,
	       int * res, int *  n, int *  m);
  void intersection (double * from0,double * from1, double * from2, 
		     double * to0,double * to1, double * to2,
		     int * n1,  int * n2);
  void num_in_interv (double * num, double * from , double * to,
		      int * m,int * n, int * res);

}

/** vec_union
 * \brief returns union of "segments" delimited by deb and fin.
 * deb and fin must be sorted vectors.
 * \author A. Lucas
 */

void vec_union (double * deb, double * fin, int * n,int * flag)
{
  int i,j;
  double curfin;
  j=0;
  curfin = fin[0];
  for(i=0; i<n[0];i++)
    {
      if (deb[i] <= curfin)
	  /* Element i is in "cluster" j: update cluster j */
	curfin = fin[j] = max( curfin, fin[i]);
      else
	{
	  /* New "cluster" j */
	  j++;
	  curfin = fin[j] = fin[i];
	  deb[j] = deb[i];
	}
      flag[i] = j+1;
    }
  n[0] = j;
}



/**
 * vec_minimal
 * \brief This function returns the maxial common segment when two 
 * segments overlap.
 * Example, with deb= 1, 15,17,100
 *               fin= 40,45,38,120
 * Function returns
 *               deb=17,100
 *               fin=38,120
 *               n=2
 * \author A. Lucas
 */


void vec_minimal (double * deb, double * fin,  int * n)
{
  int i,j;
  double curfin;
  j=0;
  
  curfin = fin[0];
  for(i=0; i<n[0];i++)
    {
      if (deb[i] <= curfin)
	{
	  /* Element i is in "cluster" j: update cluster j */
	  deb[j] = max ( deb[j],deb[i] );
	  fin[j] = min ( fin[j],fin[i] );
	  curfin = max( curfin, fin[i]);
	}
      else
	{
	  /* New "cluster" j */
	  j++;
	  curfin = fin[j] = fin[i];
	  deb[j] = deb[i];
	}
    }
  n[0] = j;
}



/** \fn  relist (int *  envdeb, int * envfin, int * posdeb, int * posfin,
	     int * res, int *  n, int *  m)
 * \brief relist put specific pos into its specific list
 * \param envdeb,envfin begining and ending positions of enveloppe
 * \param posdeb,posfin begining and ending positions of positions
 * \param res vector result (size m)
 * \param n size of envdeb
 * \param m size of posdeb
 * \author A. Lucas
 */
void relist (int *  envdeb, int * envfin, int * posdeb, int * posfin,
	     int * res, int *  n, int *  m)
{

  int i,j=0;

  for(i = 0; i< n[0]; i++)
    {
      
      for(j=j; j< m[0]; j++)
	{
	  /* we are too far in pos...*/
	  if(posdeb[j] > envfin[i])
	    break;

	  /* we are too far in env... */
	  if(posdeb[j] < envdeb[i])
	    continue;

	  /* We found a match !*/
	  if( posfin[j] <= envfin[i] )
	    res[j] = i+1; /* R indices begins with 1 (not 0) */

	}
    }


  return;
}

/** intersection
 * \brief compute intersection of segments delimited by from and to
 * (two sets of segments) 
 * \author A. Lucas
 * \param from1, from2, to1, to2 limits of (input) segments 
 * \param from0, to0 limits of (output) segments i.e. intersection.
 * \param n1, n2 length of sets of segments
 * \author A. Lucas.
 */
void intersection (double * from0,double * from1, double * from2, double * to0,
		   double * to1, double * to2,
		   int * n1,  int * n2)
{


  int i,j,k;

  /* i, sur tous les segments "1" 
   * on regarde  si un segment "2" est dedant
   *
   */
  j=1;
  k=0;
  for(i=0; i<(*n1); i++)
    {
      /*      printf("i %d\n",i);*/
      j--;
      /* new i: we go back for the 1st j: */
      while( ( j>0 ) && (to2[j] >= from1[i]))
	j--;


      /*      printf("j redescendu: %d\n",j);*/
      /* We look for the 1st segment j inside i */
      while((to2[j] <= from1[i]) && j < (*n2))
	  j++;

      
      /*      printf("j remonte: %d\n",j);*/

      for(j=j; j<(*n2);j++)
	{
	  /*  printf("j testeded: %d for i: %d\n",j,i);*/
	  /* case: no overlap for i, stop*/
	  if(from2[j] > to1[i])
	      break;

	  /* We pass all so we found intersection k */ 
	  from0[k] = max(from1[i],from2[j]);
	  to0[k] = min(to1[i],to2[j]);
	  k++;
	}

 
    }
  (*n1)=k;
  
}


/** in / match
 * \brief return indice of segment that includes num (segments must be
 * sorted and not overlapping)
 *
 * \author A. Lucas
 * \param from, to limits of segments 
 * \param num number to test (input) 
 * \param m length of num
 * \param n length of segments (from & to)
 * \param res output value is either a number (index of segment
 * found) or 0 if not found.
 *
 * \author A. Lucas.
 */
void num_in_interv (double * num, double * from , double * to,
		    int * m,int * n, int * res)
{
  int i;
  int lowerindice,upperindice;
  int indice;

  for(i =0; i < *m; i++)
    {
      res[i]=0;
      indice = (int) ((*n) /2);
      lowerindice = 0;
      upperindice = *n;

      while (( upperindice-lowerindice) > 0)
      {
	//	printf("%d %d %d %d\n",i,lowerindice,indice,upperindice);
	/*  here: we must increase indice ! */
	if(num[i] > to[indice])
	  {
	    if(lowerindice == indice)
	      lowerindice = indice +1;
	    else
	      lowerindice = indice;

	    indice += (int) ((upperindice - indice + 1) /2);
	    if(indice == lowerindice)
	      break;
	    continue;
	  }

	/*  here: we must decrease indice ! */
	if(num[i] < from[indice])
	  {
	    if(upperindice == indice)
	      upperindice = indice -1;
	    else
	      upperindice = indice;

	    indice -= (int) ((indice - lowerindice + 1 ) /2);
	    if(indice == upperindice)
	      break;
	    continue;
	  }

	/* last case: we found */
	//	if( (num[i] >= from[indice] ) && (num[i] <= to[indice]) )
	//  {
	res[i] = indice+1;
	break;
	    //}
      }
    }
}
